#define MOZ_UNIFIED_BUILD
#include "/build/js140/src/firefox-140.13.0/js/src/wasm/WasmProcess.cpp"
#ifdef PL_ARENA_CONST_ALIGN_MASK
#error "/build/js140/src/firefox-140.13.0/js/src/wasm/WasmProcess.cpp uses PL_ARENA_CONST_ALIGN_MASK, so it cannot be built in unified mode."
#undef PL_ARENA_CONST_ALIGN_MASK
#endif
#ifdef INITGUID
#error "/build/js140/src/firefox-140.13.0/js/src/wasm/WasmProcess.cpp defines INITGUID, so it cannot be built in unified mode."
#undef INITGUID
#endif
#include "/build/js140/src/firefox-140.13.0/js/src/wasm/WasmRealm.cpp"
#ifdef PL_ARENA_CONST_ALIGN_MASK
#error "/build/js140/src/firefox-140.13.0/js/src/wasm/WasmRealm.cpp uses PL_ARENA_CONST_ALIGN_MASK, so it cannot be built in unified mode."
#undef PL_ARENA_CONST_ALIGN_MASK
#endif
#ifdef INITGUID
#error "/build/js140/src/firefox-140.13.0/js/src/wasm/WasmRealm.cpp defines INITGUID, so it cannot be built in unified mode."
#undef INITGUID
#endif
#include "/build/js140/src/firefox-140.13.0/js/src/wasm/WasmSerialize.cpp"
#ifdef PL_ARENA_CONST_ALIGN_MASK
#error "/build/js140/src/firefox-140.13.0/js/src/wasm/WasmSerialize.cpp uses PL_ARENA_CONST_ALIGN_MASK, so it cannot be built in unified mode."
#undef PL_ARENA_CONST_ALIGN_MASK
#endif
#ifdef INITGUID
#error "/build/js140/src/firefox-140.13.0/js/src/wasm/WasmSerialize.cpp defines INITGUID, so it cannot be built in unified mode."
#undef INITGUID
#endif
#include "/build/js140/src/firefox-140.13.0/js/src/wasm/WasmSignalHandlers.cpp"
#ifdef PL_ARENA_CONST_ALIGN_MASK
#error "/build/js140/src/firefox-140.13.0/js/src/wasm/WasmSignalHandlers.cpp uses PL_ARENA_CONST_ALIGN_MASK, so it cannot be built in unified mode."
#undef PL_ARENA_CONST_ALIGN_MASK
#endif
#ifdef INITGUID
#error "/build/js140/src/firefox-140.13.0/js/src/wasm/WasmSignalHandlers.cpp defines INITGUID, so it cannot be built in unified mode."
#undef INITGUID
#endif
#include "/build/js140/src/firefox-140.13.0/js/src/wasm/WasmStaticTypeDefs.cpp"
#ifdef PL_ARENA_CONST_ALIGN_MASK
#error "/build/js140/src/firefox-140.13.0/js/src/wasm/WasmStaticTypeDefs.cpp uses PL_ARENA_CONST_ALIGN_MASK, so it cannot be built in unified mode."
#undef PL_ARENA_CONST_ALIGN_MASK
#endif
#ifdef INITGUID
#error "/build/js140/src/firefox-140.13.0/js/src/wasm/WasmStaticTypeDefs.cpp defines INITGUID, so it cannot be built in unified mode."
#undef INITGUID
#endif
#include "/build/js140/src/firefox-140.13.0/js/src/wasm/WasmStubs.cpp"
#ifdef PL_ARENA_CONST_ALIGN_MASK
#error "/build/js140/src/firefox-140.13.0/js/src/wasm/WasmStubs.cpp uses PL_ARENA_CONST_ALIGN_MASK, so it cannot be built in unified mode."
#undef PL_ARENA_CONST_ALIGN_MASK
#endif
#ifdef INITGUID
#error "/build/js140/src/firefox-140.13.0/js/src/wasm/WasmStubs.cpp defines INITGUID, so it cannot be built in unified mode."
#undef INITGUID
#endif