#define MOZ_UNIFIED_BUILD
#include "/build/js140/src/firefox-140.13.0/js/src/wasm/WasmBuiltinModule.cpp"
#ifdef PL_ARENA_CONST_ALIGN_MASK
#error "/build/js140/src/firefox-140.13.0/js/src/wasm/WasmBuiltinModule.cpp uses PL_ARENA_CONST_ALIGN_MASK, so it cannot be built in unified mode."
#undef PL_ARENA_CONST_ALIGN_MASK
#endif
#ifdef INITGUID
#error "/build/js140/src/firefox-140.13.0/js/src/wasm/WasmBuiltinModule.cpp defines INITGUID, so it cannot be built in unified mode."
#undef INITGUID
#endif
#include "/build/js140/src/firefox-140.13.0/js/src/wasm/WasmBuiltins.cpp"
#ifdef PL_ARENA_CONST_ALIGN_MASK
#error "/build/js140/src/firefox-140.13.0/js/src/wasm/WasmBuiltins.cpp uses PL_ARENA_CONST_ALIGN_MASK, so it cannot be built in unified mode."
#undef PL_ARENA_CONST_ALIGN_MASK
#endif
#ifdef INITGUID
#error "/build/js140/src/firefox-140.13.0/js/src/wasm/WasmBuiltins.cpp defines INITGUID, so it cannot be built in unified mode."
#undef INITGUID
#endif
#include "/build/js140/src/firefox-140.13.0/js/src/wasm/WasmCode.cpp"
#ifdef PL_ARENA_CONST_ALIGN_MASK
#error "/build/js140/src/firefox-140.13.0/js/src/wasm/WasmCode.cpp uses PL_ARENA_CONST_ALIGN_MASK, so it cannot be built in unified mode."
#undef PL_ARENA_CONST_ALIGN_MASK
#endif
#ifdef INITGUID
#error "/build/js140/src/firefox-140.13.0/js/src/wasm/WasmCode.cpp defines INITGUID, so it cannot be built in unified mode."
#undef INITGUID
#endif
#include "/build/js140/src/firefox-140.13.0/js/src/wasm/WasmCodegenTypes.cpp"
#ifdef PL_ARENA_CONST_ALIGN_MASK
#error "/build/js140/src/firefox-140.13.0/js/src/wasm/WasmCodegenTypes.cpp uses PL_ARENA_CONST_ALIGN_MASK, so it cannot be built in unified mode."
#undef PL_ARENA_CONST_ALIGN_MASK
#endif
#ifdef INITGUID
#error "/build/js140/src/firefox-140.13.0/js/src/wasm/WasmCodegenTypes.cpp defines INITGUID, so it cannot be built in unified mode."
#undef INITGUID
#endif
#include "/build/js140/src/firefox-140.13.0/js/src/wasm/WasmCompile.cpp"
#ifdef PL_ARENA_CONST_ALIGN_MASK
#error "/build/js140/src/firefox-140.13.0/js/src/wasm/WasmCompile.cpp uses PL_ARENA_CONST_ALIGN_MASK, so it cannot be built in unified mode."
#undef PL_ARENA_CONST_ALIGN_MASK
#endif
#ifdef INITGUID
#error "/build/js140/src/firefox-140.13.0/js/src/wasm/WasmCompile.cpp defines INITGUID, so it cannot be built in unified mode."
#undef INITGUID
#endif
#include "/build/js140/src/firefox-140.13.0/js/src/wasm/WasmDebug.cpp"
#ifdef PL_ARENA_CONST_ALIGN_MASK
#error "/build/js140/src/firefox-140.13.0/js/src/wasm/WasmDebug.cpp uses PL_ARENA_CONST_ALIGN_MASK, so it cannot be built in unified mode."
#undef PL_ARENA_CONST_ALIGN_MASK
#endif
#ifdef INITGUID
#error "/build/js140/src/firefox-140.13.0/js/src/wasm/WasmDebug.cpp defines INITGUID, so it cannot be built in unified mode."
#undef INITGUID
#endif