#define MOZ_UNIFIED_BUILD
#include "/build/js140/src/firefox-140.13.0/js/src/jit/JitContext.cpp"
#ifdef PL_ARENA_CONST_ALIGN_MASK
#error "/build/js140/src/firefox-140.13.0/js/src/jit/JitContext.cpp uses PL_ARENA_CONST_ALIGN_MASK, so it cannot be built in unified mode."
#undef PL_ARENA_CONST_ALIGN_MASK
#endif
#ifdef INITGUID
#error "/build/js140/src/firefox-140.13.0/js/src/jit/JitContext.cpp defines INITGUID, so it cannot be built in unified mode."
#undef INITGUID
#endif
#include "/build/js140/src/firefox-140.13.0/js/src/jit/JitFrames.cpp"
#ifdef PL_ARENA_CONST_ALIGN_MASK
#error "/build/js140/src/firefox-140.13.0/js/src/jit/JitFrames.cpp uses PL_ARENA_CONST_ALIGN_MASK, so it cannot be built in unified mode."
#undef PL_ARENA_CONST_ALIGN_MASK
#endif
#ifdef INITGUID
#error "/build/js140/src/firefox-140.13.0/js/src/jit/JitFrames.cpp defines INITGUID, so it cannot be built in unified mode."
#undef INITGUID
#endif
#include "/build/js140/src/firefox-140.13.0/js/src/jit/JitHints.cpp"
#ifdef PL_ARENA_CONST_ALIGN_MASK
#error "/build/js140/src/firefox-140.13.0/js/src/jit/JitHints.cpp uses PL_ARENA_CONST_ALIGN_MASK, so it cannot be built in unified mode."
#undef PL_ARENA_CONST_ALIGN_MASK
#endif
#ifdef INITGUID
#error "/build/js140/src/firefox-140.13.0/js/src/jit/JitHints.cpp defines INITGUID, so it cannot be built in unified mode."
#undef INITGUID
#endif
#include "/build/js140/src/firefox-140.13.0/js/src/jit/JitOptions.cpp"
#ifdef PL_ARENA_CONST_ALIGN_MASK
#error "/build/js140/src/firefox-140.13.0/js/src/jit/JitOptions.cpp uses PL_ARENA_CONST_ALIGN_MASK, so it cannot be built in unified mode."
#undef PL_ARENA_CONST_ALIGN_MASK
#endif
#ifdef INITGUID
#error "/build/js140/src/firefox-140.13.0/js/src/jit/JitOptions.cpp defines INITGUID, so it cannot be built in unified mode."
#undef INITGUID
#endif
#include "/build/js140/src/firefox-140.13.0/js/src/jit/JitScript.cpp"
#ifdef PL_ARENA_CONST_ALIGN_MASK
#error "/build/js140/src/firefox-140.13.0/js/src/jit/JitScript.cpp uses PL_ARENA_CONST_ALIGN_MASK, so it cannot be built in unified mode."
#undef PL_ARENA_CONST_ALIGN_MASK
#endif
#ifdef INITGUID
#error "/build/js140/src/firefox-140.13.0/js/src/jit/JitScript.cpp defines INITGUID, so it cannot be built in unified mode."
#undef INITGUID
#endif
#include "/build/js140/src/firefox-140.13.0/js/src/jit/JitSpewer.cpp"
#ifdef PL_ARENA_CONST_ALIGN_MASK
#error "/build/js140/src/firefox-140.13.0/js/src/jit/JitSpewer.cpp uses PL_ARENA_CONST_ALIGN_MASK, so it cannot be built in unified mode."
#undef PL_ARENA_CONST_ALIGN_MASK
#endif
#ifdef INITGUID
#error "/build/js140/src/firefox-140.13.0/js/src/jit/JitSpewer.cpp defines INITGUID, so it cannot be built in unified mode."
#undef INITGUID
#endif