#define MOZ_UNIFIED_BUILD
#include "/build/js140/src/firefox-140.13.0/mfbt/Assertions.cpp"
#ifdef PL_ARENA_CONST_ALIGN_MASK
#error "/build/js140/src/firefox-140.13.0/mfbt/Assertions.cpp uses PL_ARENA_CONST_ALIGN_MASK, so it cannot be built in unified mode."
#undef PL_ARENA_CONST_ALIGN_MASK
#endif
#ifdef INITGUID
#error "/build/js140/src/firefox-140.13.0/mfbt/Assertions.cpp defines INITGUID, so it cannot be built in unified mode."
#undef INITGUID
#endif
#include "/build/js140/src/firefox-140.13.0/mfbt/ChaosMode.cpp"
#ifdef PL_ARENA_CONST_ALIGN_MASK
#error "/build/js140/src/firefox-140.13.0/mfbt/ChaosMode.cpp uses PL_ARENA_CONST_ALIGN_MASK, so it cannot be built in unified mode."
#undef PL_ARENA_CONST_ALIGN_MASK
#endif
#ifdef INITGUID
#error "/build/js140/src/firefox-140.13.0/mfbt/ChaosMode.cpp defines INITGUID, so it cannot be built in unified mode."
#undef INITGUID
#endif
#include "/build/js140/src/firefox-140.13.0/mfbt/FloatingPoint.cpp"
#ifdef PL_ARENA_CONST_ALIGN_MASK
#error "/build/js140/src/firefox-140.13.0/mfbt/FloatingPoint.cpp uses PL_ARENA_CONST_ALIGN_MASK, so it cannot be built in unified mode."
#undef PL_ARENA_CONST_ALIGN_MASK
#endif
#ifdef INITGUID
#error "/build/js140/src/firefox-140.13.0/mfbt/FloatingPoint.cpp defines INITGUID, so it cannot be built in unified mode."
#undef INITGUID
#endif
#include "/build/js140/src/firefox-140.13.0/mfbt/HashFunctions.cpp"
#ifdef PL_ARENA_CONST_ALIGN_MASK
#error "/build/js140/src/firefox-140.13.0/mfbt/HashFunctions.cpp uses PL_ARENA_CONST_ALIGN_MASK, so it cannot be built in unified mode."
#undef PL_ARENA_CONST_ALIGN_MASK
#endif
#ifdef INITGUID
#error "/build/js140/src/firefox-140.13.0/mfbt/HashFunctions.cpp defines INITGUID, so it cannot be built in unified mode."
#undef INITGUID
#endif
#include "/build/js140/src/firefox-140.13.0/mfbt/JSONWriter.cpp"
#ifdef PL_ARENA_CONST_ALIGN_MASK
#error "/build/js140/src/firefox-140.13.0/mfbt/JSONWriter.cpp uses PL_ARENA_CONST_ALIGN_MASK, so it cannot be built in unified mode."
#undef PL_ARENA_CONST_ALIGN_MASK
#endif
#ifdef INITGUID
#error "/build/js140/src/firefox-140.13.0/mfbt/JSONWriter.cpp defines INITGUID, so it cannot be built in unified mode."
#undef INITGUID
#endif
#include "/build/js140/src/firefox-140.13.0/mfbt/Poison.cpp"
#ifdef PL_ARENA_CONST_ALIGN_MASK
#error "/build/js140/src/firefox-140.13.0/mfbt/Poison.cpp uses PL_ARENA_CONST_ALIGN_MASK, so it cannot be built in unified mode."
#undef PL_ARENA_CONST_ALIGN_MASK
#endif
#ifdef INITGUID
#error "/build/js140/src/firefox-140.13.0/mfbt/Poison.cpp defines INITGUID, so it cannot be built in unified mode."
#undef INITGUID
#endif
#include "/build/js140/src/firefox-140.13.0/mfbt/RandomNum.cpp"
#ifdef PL_ARENA_CONST_ALIGN_MASK
#error "/build/js140/src/firefox-140.13.0/mfbt/RandomNum.cpp uses PL_ARENA_CONST_ALIGN_MASK, so it cannot be built in unified mode."
#undef PL_ARENA_CONST_ALIGN_MASK
#endif
#ifdef INITGUID
#error "/build/js140/src/firefox-140.13.0/mfbt/RandomNum.cpp defines INITGUID, so it cannot be built in unified mode."
#undef INITGUID
#endif
#include "/build/js140/src/firefox-140.13.0/mfbt/RefCounted.cpp"
#ifdef PL_ARENA_CONST_ALIGN_MASK
#error "/build/js140/src/firefox-140.13.0/mfbt/RefCounted.cpp uses PL_ARENA_CONST_ALIGN_MASK, so it cannot be built in unified mode."
#undef PL_ARENA_CONST_ALIGN_MASK
#endif
#ifdef INITGUID
#error "/build/js140/src/firefox-140.13.0/mfbt/RefCounted.cpp defines INITGUID, so it cannot be built in unified mode."
#undef INITGUID
#endif
#include "/build/js140/src/firefox-140.13.0/mfbt/SHA1.cpp"
#ifdef PL_ARENA_CONST_ALIGN_MASK
#error "/build/js140/src/firefox-140.13.0/mfbt/SHA1.cpp uses PL_ARENA_CONST_ALIGN_MASK, so it cannot be built in unified mode."
#undef PL_ARENA_CONST_ALIGN_MASK
#endif
#ifdef INITGUID
#error "/build/js140/src/firefox-140.13.0/mfbt/SHA1.cpp defines INITGUID, so it cannot be built in unified mode."
#undef INITGUID
#endif
#include "/build/js140/src/firefox-140.13.0/mfbt/TaggedAnonymousMemory.cpp"
#ifdef PL_ARENA_CONST_ALIGN_MASK
#error "/build/js140/src/firefox-140.13.0/mfbt/TaggedAnonymousMemory.cpp uses PL_ARENA_CONST_ALIGN_MASK, so it cannot be built in unified mode."
#undef PL_ARENA_CONST_ALIGN_MASK
#endif
#ifdef INITGUID
#error "/build/js140/src/firefox-140.13.0/mfbt/TaggedAnonymousMemory.cpp defines INITGUID, so it cannot be built in unified mode."
#undef INITGUID
#endif
#include "/build/js140/src/firefox-140.13.0/mfbt/UniquePtrExtensions.cpp"
#ifdef PL_ARENA_CONST_ALIGN_MASK
#error "/build/js140/src/firefox-140.13.0/mfbt/UniquePtrExtensions.cpp uses PL_ARENA_CONST_ALIGN_MASK, so it cannot be built in unified mode."
#undef PL_ARENA_CONST_ALIGN_MASK
#endif
#ifdef INITGUID
#error "/build/js140/src/firefox-140.13.0/mfbt/UniquePtrExtensions.cpp defines INITGUID, so it cannot be built in unified mode."
#undef INITGUID
#endif
#include "/build/js140/src/firefox-140.13.0/mfbt/Unused.cpp"
#ifdef PL_ARENA_CONST_ALIGN_MASK
#error "/build/js140/src/firefox-140.13.0/mfbt/Unused.cpp uses PL_ARENA_CONST_ALIGN_MASK, so it cannot be built in unified mode."
#undef PL_ARENA_CONST_ALIGN_MASK
#endif
#ifdef INITGUID
#error "/build/js140/src/firefox-140.13.0/mfbt/Unused.cpp defines INITGUID, so it cannot be built in unified mode."
#undef INITGUID
#endif
#include "/build/js140/src/firefox-140.13.0/mfbt/Utf8.cpp"
#ifdef PL_ARENA_CONST_ALIGN_MASK
#error "/build/js140/src/firefox-140.13.0/mfbt/Utf8.cpp uses PL_ARENA_CONST_ALIGN_MASK, so it cannot be built in unified mode."
#undef PL_ARENA_CONST_ALIGN_MASK
#endif
#ifdef INITGUID
#error "/build/js140/src/firefox-140.13.0/mfbt/Utf8.cpp defines INITGUID, so it cannot be built in unified mode."
#undef INITGUID
#endif
#include "/build/js140/src/firefox-140.13.0/mfbt/double-conversion/double-conversion/bignum-dtoa.cc"
#ifdef PL_ARENA_CONST_ALIGN_MASK
#error "/build/js140/src/firefox-140.13.0/mfbt/double-conversion/double-conversion/bignum-dtoa.cc uses PL_ARENA_CONST_ALIGN_MASK, so it cannot be built in unified mode."
#undef PL_ARENA_CONST_ALIGN_MASK
#endif
#ifdef INITGUID
#error "/build/js140/src/firefox-140.13.0/mfbt/double-conversion/double-conversion/bignum-dtoa.cc defines INITGUID, so it cannot be built in unified mode."
#undef INITGUID
#endif
#include "/build/js140/src/firefox-140.13.0/mfbt/double-conversion/double-conversion/bignum.cc"
#ifdef PL_ARENA_CONST_ALIGN_MASK
#error "/build/js140/src/firefox-140.13.0/mfbt/double-conversion/double-conversion/bignum.cc uses PL_ARENA_CONST_ALIGN_MASK, so it cannot be built in unified mode."
#undef PL_ARENA_CONST_ALIGN_MASK
#endif
#ifdef INITGUID
#error "/build/js140/src/firefox-140.13.0/mfbt/double-conversion/double-conversion/bignum.cc defines INITGUID, so it cannot be built in unified mode."
#undef INITGUID
#endif
#include "/build/js140/src/firefox-140.13.0/mfbt/double-conversion/double-conversion/cached-powers.cc"
#ifdef PL_ARENA_CONST_ALIGN_MASK
#error "/build/js140/src/firefox-140.13.0/mfbt/double-conversion/double-conversion/cached-powers.cc uses PL_ARENA_CONST_ALIGN_MASK, so it cannot be built in unified mode."
#undef PL_ARENA_CONST_ALIGN_MASK
#endif
#ifdef INITGUID
#error "/build/js140/src/firefox-140.13.0/mfbt/double-conversion/double-conversion/cached-powers.cc defines INITGUID, so it cannot be built in unified mode."
#undef INITGUID
#endif